#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_XML_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <com/maroontress/libcgi/db/Box.h>
#include <com/maroontress/libcgi/db/XML.h>
#include <com/maroontress/libcgi/util/Tub.h>

struct Box {
    char *key;
    void *data;
    size_t size;
    char **names;
    int namesLen;
};

void
Box_free(struct Box *b)
{
    if (b == NULL)
	return;
    free(b->key);
    free(b->data);
    free(b->names);
    free(b);
}

static int
validate(const char *data, size_t size)
{
    size_t n, m, valSize, len;
    const char *valData;

    len = 0;
    while (size > 0) {
	if (size < sizeof(n))
	    return -1;
        bcopy(data, &n, sizeof(n));
        data += sizeof(n);
	size -= sizeof(n);

	if (size < n || memchr(data, 0, n) == NULL)
	    return -1;
	m = strlen(data) + 1;
	valData = data + m;
	valSize = size - m;
	switch (*valData) {
	default:
	    return -1;
	case 'i':
	    if (valSize < 1 + sizeof(int))
		return -1;
	    break;
	case 's':
            if (valSize < 1 || memchr(valData + 1, 0, valSize - 1) == NULL)
		return -1;
	    break;
	}

        data += n;
        size -= n;
	++len;
    }
    return len;
}

struct Box *
Box_create(const char *key, const void *data, size_t size)
{
    struct Box *b;

    if ((b = (struct Box *)malloc(sizeof(*b))) == NULL)
	return NULL;
    b->key = NULL;
    b->data = NULL;
    b->size = size;
    b->names = NULL;
    b->namesLen = 0;

    if ((b->key = strdup(key)) == NULL
	|| (b->data = malloc(size)) == NULL
	|| (b->namesLen = validate(data, size)) < 0) {
	Box_free(b);
	return NULL;
    }
    bcopy(data, b->data, size);
    return b;
}

static int
finds(struct Box *b, const char *name, /*@out@*/ char **value)
{
    char *data = b->data;
    size_t size = b->size, n;

    while (size > 0) {
        bcopy(data, &n, sizeof(n));
        data += sizeof(n);
	size -= sizeof(n);

        if (strcmp(name, data) == 0) {
	    size_t m = strlen(data) + 1;
            *value = data + m;
	    return 1;
	}

        data += n;
        size -= n;
    }
    return 0;
}

int
Box_getInt(struct Box *b, const char *name, /*@out@*/ int *value)
{
    char *data;

    if (!finds(b, name, &data))
	return -1;
    if (*data != 'i')
	return -1;
    bcopy(data + 1, value, sizeof(*value));
    return 0;
}

int
Box_getString(struct Box *b, const char *name, /*@out@*/ char **value)
{
    char *data;

    if (!finds(b, name, &data))
	return -1;
    if (*data != 's')
	return -1;
    *value = data + 1;
    return 0;
}

char *
Box_getKey(struct Box *b)
{
    return b->key;
}

char **
Box_getNames(struct Box *b)
{
    char **names;
    char *data = b->data;
    size_t size = b->size, n;
    int k;

    if (b->names != NULL)
	return b->names;
    if ((names = (char **)malloc(sizeof(char *) * (b->namesLen + 1))) == NULL)
	return NULL;
    for (k = 0; size > 0 && k < b->namesLen; ++k) {
        bcopy(data, &n, sizeof(n));
        data += sizeof(n);
	size -= sizeof(n);

	names[k] = data;

        data += n;
        size -= n;
    }
    names[k] = NULL;
    b->names = names;
    return b->names;
}

int
Box_getType(struct Box *b, const char *name, enum Box_Type *type)
{
    char *data;

    if (!finds(b, name, &data))
	return -1;
    switch (*data) {
    default:
	return -1;
    case 'i':
	*type = Box_Type_INT;
	break;
    case 's':
	*type = Box_Type_STRING;
	break;
    }
    return 0;
}

int
Box_writeKeyXMLEscapedToTub(struct Box *box, struct Tub *tub)
{
    return XML_writeEscapedToTub(tub, Box_getKey(box));
}

static int
writeToTub(struct Box *box, const char *name, struct Tub *tub,
	   int (*writer)(struct Tub *, const char *))
{
    enum Box_Type type;
    char *strValue, buf[16];
    int intValue;

    if (Box_getType(box, name, &type) < 0)
	return -1;
    switch (type) {
    case Box_Type_INT:
	if (Box_getInt(box, name, &intValue) < 0)
	    return -1;
	snprintf(buf, sizeof(buf), "%d", intValue);
	if (Tub_writeCString(tub, buf) < 0)
	    return -1;
	break;
    case Box_Type_STRING:
	if (Box_getString(box, name, &strValue) < 0
	    || writer(tub, strValue) < 0)
	    return -1;
	break;
    }
    return 0;
}

int
Box_writeToTub(struct Box *box, const char *name, struct Tub *tub)
{
    return writeToTub(box, name, tub, Tub_writeCString);
}

int
Box_writeXMLEscapedToTub(struct Box *box, const char *name, struct Tub *tub)
{
    return writeToTub(box, name, tub, XML_writeEscapedToTub);
}

static int
writeXMLElementToTub(struct Box *box, struct Tub *tub,
		     const char *elementName, va_list ap)
{
    const char *name;

    if (Tub_writeByte(tub, '<') < 0
	|| Tub_writeCString(tub, elementName) < 0
	|| Tub_writeByte(tub, ' ') < 0) {
	return -1;
    }
    while ((name = va_arg(ap, const char *)) != NULL) {
	if (Tub_writeCString(tub, name) < 0
	    || Tub_writeCString(tub, "=\"") < 0
	    || Box_writeXMLEscapedToTub(box, name, tub) < 0
	    || Tub_writeCString(tub, "\" ") < 0) {
	    return -1;
	}
    }
    if (Tub_writeCString(tub, "/>") < 0) {
	return -1;
    }
    return 0;
}

int
Box_writeXMLElementToTub(struct Box *box, struct Tub *tub,
			 const char *elementName, ...)
{
    va_list ap;
    int s;

    va_start(ap, elementName);
    s = writeXMLElementToTub(box, tub, elementName, ap);
    va_end(ap);
    return s;
}

static int
writeXMLElementWithKeyToTub(struct Box *box, struct Tub *tub,
			    const char *elementName, const char *keyAttrName,
			    va_list ap)
{
    const char *name;

    if (Tub_writeByte(tub, '<') < 0
	|| Tub_writeCString(tub, elementName) < 0
	|| Tub_writeByte(tub, ' ') < 0
	|| Tub_writeCString(tub, keyAttrName) < 0
	|| Tub_writeCString(tub, "=\"") < 0
	|| Box_writeKeyXMLEscapedToTub(box, tub) < 0
	|| Tub_writeCString(tub, "\" ") < 0) {
	return -1;
    }
    while ((name = va_arg(ap, const char *)) != NULL) {
	if (Tub_writeCString(tub, name) < 0
	    || Tub_writeCString(tub, "=\"") < 0
	    || Box_writeXMLEscapedToTub(box, name, tub) < 0
	    || Tub_writeCString(tub, "\" ") < 0) {
	    return -1;
	}
    }
    if (Tub_writeCString(tub, "/>") < 0) {
	return -1;
    }
    return 0;
}

int
Box_writeXMLElementWithKeyToTub(struct Box *box, struct Tub *tub,
				const char *elementName,
				const char *keyAttrName, ...)
{
    va_list ap;
    int s;

    va_start(ap, keyAttrName);
    s = writeXMLElementWithKeyToTub(box, tub, elementName, keyAttrName, ap);
    va_end(ap);
    return s;
}

char *
Box_createXMLElement(struct Box *box, const char *elementName, ...)
{
    va_list ap;
    struct Tub *tub;
    char *e;

    if ((tub = Tub_create()) == NULL)
	return NULL;
    va_start(ap, elementName);
    e = (writeXMLElementToTub(box, tub, elementName, ap) < 0
	 || Tub_writeByte(tub, 0) < 0) ? NULL : Tub_createByteArray(tub);
    va_end(ap);
    Tub_free(tub);
    return e;
}

char *
Box_createXMLElementWithKey(struct Box *box, const char *elementName,
			    const char *keyAttrName, ...)
{
    va_list ap;
    struct Tub *tub;
    char *e;

    if ((tub = Tub_create()) == NULL)
	return NULL;
    va_start(ap, keyAttrName);
    e = (writeXMLElementWithKeyToTub(box, tub, elementName, keyAttrName,
				     ap) < 0
	 || Tub_writeByte(tub, 0) < 0) ? NULL : Tub_createByteArray(tub);
    va_end(ap);
    Tub_free(tub);
    return e;
}
